import json
import logging
import os
import re

from channels import layers
from django.shortcuts import render, redirect

from reface import const
from reface.core import errors
from reface.dashboard import forms
from reface.dashboard.models import Box, Boxes
from resh.protocols.dali_utils.external_commands.dali2 import events

FIRMWARE_DIR = '/tmp/rapida/firmwares/'
BUILDS_DIR = '/tmp/rapida/builds/'

logger = logging.getLogger('reface')

with open(os.environ.get('REFACE_CONFIG'), 'r') as file:
    config = json.loads(file.read())

if 'dump' in config:
    boxes = Boxes(config.get('dump'))
    boxes.init()
else:
    logger.warning('Dump file not set! Break...')


def root(request):
    if not boxes.configured:
        return render(request, 'tools/configurator.html')
    if boxes.is_sw_gateway:
        return redirect('/rapida/0')
    return render(request, 'objects/system.html', context=dict(name=boxes.name,
                                                               boxes=sorted(boxes.values(), key=lambda box: box.id)))


def upgrade(request):
    if request.method == 'POST':
        has_dependencies = False
        build = handle_uploaded_file(BUILDS_DIR, request.FILES['build'])
        for fs in request.FILES.getlist('dependencies', []):
            has_dependencies = True
            handle_uploaded_file(BUILDS_DIR, fs)

        cmd = 'sudo pip3 install %s --upgrade' % build
        if has_dependencies:
            cmd += ' --no-index --find-links=%s' % BUILDS_DIR

        os.system(cmd)
        os.system('sudo supervisorctl restart rapida_reface')

        return redirect('/rapida/')
    else:
        return render(request, 'tools/version_upgrade.html')


def redirect_page(request, url, timeout):
    return render(request, 'tools/redirector.html', context=dict(url=url, timeout=timeout))


def configure(request):
    boxes.set_type(request.GET['type'])
    return redirect('/rapida/')


def reload(request):
    boxes.init()
    return redirect('/rapida/')


def controller(request):
    if 'update' in request.GET:
        if boxes.has_controller:
            boxes.controller.renew()
    elif 'disconnect' in request.GET:
        boxes.disconnect_sw_gateway()
        return redirect('/rapida/')
    elif 'upgrade' in request.GET:
        if 'time' in request.GET:
            boxes.controller.set_time(request.GET['time'])
        if 'date' in request.GET:
            boxes.controller.set_date(request.GET['date'])
        if 'timezone' in request.GET:
            boxes.controller.set_timezone(request.GET['timezone'])
        if 'dns0' in request.GET or 'dns1' in request.GET:
            boxes.controller.set_dns(' '.join([request.GET.get('dns0', ''), request.GET.get('dns1', '')]).strip().split())
        if 'ip' in request.GET or 'mask' in request.GET or 'gateway' in request.GET:
            boxes.controller.set_ip(ip=request.GET.get('ip'), mask=request.GET.get('mask'), gateway=request.GET.get('gateway'))

    if request.GET:
        return redirect(request.path)

    return render(request, 'objects/controller.html', context=dict(name=boxes.name,
                                                                   is_sw_gateway=boxes.is_sw_gateway,
                                                                   controller=boxes.controller))


def set_box_name(request, box_id):
    boxes.set_box_name(int(box_id), request.GET.get('name', 'unknown'))
    return redirect(request.GET.get('redirect', '/rapida/'))


def set_name(request):
    boxes.set_name(request.GET.get('name', 'unknown'))
    return redirect(request.path[:request.path.index('set_name')])


def add_box(request):
    occupied, index = boxes.get_indexes()
    action = request.GET.get('action', None)
    index = int(request.GET.get('index', index))
    host = re.sub(r'[\d+][:\d+]*$', '', request.get_host().split(':')[0])
    host = str(request.GET.get('host', host))
    port = int(request.GET.get('port', Box.DEF_PORT))
    name = request.GET.get('name', '')
    role = Box.FAKE

    box = Box(index, host=host, port=port, name=name, role=role)

    if action == 'add':
        box.role = Box.GATEWAY
        boxes.set(box)
        return redirect(f'/rapida/box/{index}')
    elif action == 'conf':
        box.role = Box.GATEWAY
        boxes[index].status = Box.CONNECTED
        boxes.flush()
        return redirect(f'/rapida/box/{index}')
    elif action == 'remove':
        boxes.remove(index)
        return redirect('/rapida/')
    elif action == 'commit':
        boxes.set_box_name(index, name)
        return redirect(f'/rapida/box/{index}')

    return render(request, 'objects/box.html', context=dict(name=boxes.name, box=box, occupied=occupied))


def box(request, box_id):
    box = boxes.get(int(box_id))
    if 'update' in request.GET:
        boxes.update_box(box)
        return redirect(request.path)
    return render(request, 'objects/box.html', context=dict(name=boxes.name, box=box, occupied=boxes.get_indexes()))


def channel(request, box_id, module_id, channel_id):
    try:
        _module = [m for m in boxes[int(box_id)].modules if m['address'] == int(module_id)][0]
    except IndexError:
        logging.error('module not found')
    else:
        return render(
            request, 'objects/channel.html',
            context=dict(
                name=boxes.name,
                box=int(box_id),
                box_name=boxes[int(box_id)].name,
                module=int(module_id),
                channel=int(channel_id),
                channels=(
                    ch['id']
                    for ch in _module.get('channels', [])
                    if ch.get('powered', False)
                )
            )
        )


def line_builder(request, box_id, module_id, channel_id):
    form = forms.LineBuilder(*(request.POST,) if request.method == 'POST' else ())
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    if request.method == 'POST':
        if 'last_addressing' in request.POST:
            return redirect(request.path + 'process/')
        elif form.is_valid():
            try:
                command = form.command(request.POST.get('action'))
                try:
                    layers.get_channel_layer(const.LINE_BUILDER).build_line(
                        command, box_id, module_id, channel_id, form.mode, form.start_address
                    )
                except errors.AlreadyStarted:
                    pass
                return redirect(request.path + 'process/')
            except BaseException as ex:
                logging.error(ex)

    return render(request, 'tools/line_builder.html', context=dict(
        name=boxes.name, box=box_id, module=module_id, channel=channel_id, form=form, room=room))


def line_builder_process(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    return render(request, 'tools/line_building.html', context=dict(
        name=boxes.name, box=box_id, module=module_id, channel=channel_id, room=room,
        action_verbose=forms.LineBuilder.verbose_name(
            layers.get_channel_layer(const.LINE_BUILDER).action_name(box_id, module_id, channel_id)
        )
    ))


def address_editor(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    dali_type = request.GET.get('dali_type', forms.dali1.Dali1Protocol.domain_name)
    if dali_type in request.GET:
        addresses = [i[1:] for i in request.GET.get(dali_type).split(',')]
    else:
        addresses = range(0, 64)
    return render(
        request,
        'tools/address_editor.html',
        context=dict(
            name=boxes.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=addresses
        ))


def probation(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    return render(
        request,
        'tools/probation.html',
        context=dict(
            name=boxes.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=range(0, 64),
            commands=(
                # Real command name, Displayed name
                (forms.dali1.gear.QueryActualLevel.__name__, forms.dali1.gear.QueryActualLevel.__name__),
                (forms.dali1.gear.QueryMaxLevel.__name__, forms.dali1.gear.QueryMaxLevel.__name__),
                (forms.dali1.gear.QueryMinLevel.__name__, forms.dali1.gear.QueryMinLevel.__name__),
                (forms.dali1.led.QueryDimmingCurve.__name__, forms.dali1.led.QueryDimmingCurve.__name__),
            )
        ))


def devices(request, box_id, module_id, channel_id):
    form = forms.LineScanner(*(request.POST,) if request.method == 'POST' else ())

    return render(request, 'tools/devices.html', context=dict(
        form=form,
        room='%s/%s/%s' % (box_id, module_id, channel_id),
        name=boxes.name, box=box_id, module=module_id, channel=channel_id,
        addresses=range(0, 64),
        groups_0_15=range(0, 16),
        groups_16_31=range(16, 32),
        columns=form.columns(),
        columns_width=6,
        commands=form.commands(),
        commands_width=5,
        event_filters=events.types
    ))


def module(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    if 'upgrade' in request.POST:
        if 'address' in request.POST:
            address = int(request.POST.get('address'))
            if module_id != address:
                boxes.set_module_address(box_id, module_id, address)
                module_id = address
        if 'firmware' in request.FILES:
            firmware = handle_uploaded_file(FIRMWARE_DIR, request.FILES['firmware'])
            try:
                boxes.update_module_firmware(box_id, module_id, firmware)
            except errors.TerminalError as ex:
                logging.error(ex)
            os.remove(firmware)
        return redirect(f'/rapida/{box_id}/{module_id}')
    return render(request, 'objects/module.html', context=dict(name=boxes.name,
                                                               box=box_id,
                                                               module=boxes.module_info(box_id, module_id),
                                                               modules=boxes.get_modules(box_id)))


def handle_uploaded_file(dest, f):
    os.makedirs(dest, exist_ok=True)
    filename = os.path.join(dest, f.name)
    with open(filename, 'wb+') as dest:
        for chunk in f.chunks():
            dest.write(chunk)
    return filename
